@@ -1,10 +1,30 @@
 /* 
- * Copyright (c) 2000 QUALCOMM Incorporated.  All rights reserved.
+ * Copyright (c) 2001 QUALCOMM Incorporated.  All rights reserved.
  * The file license.txt specifies the terms for use, modification,
  * and redistribution.
  *
  * Revisions:
  *
+ *     04/03/01  [rcg]
+ *              - Be a little nicer with "-v".
+ *
+ *     02/13/01  [rcg]
+ *              - Don't abort on weird accept() errors.
+ *
+ *     02/12/01  [rcg]
+ *              - HUP now closes and reopens trace file in standalone mode.
+ *
+ *     12/21/00  [rcg]
+ *              - Fixed option recognition.
+ *              - Ignore additional accept() errors, per Stevens 5.11
+ *                (thanks to Carles Munyoz for reporting this).
+ *
+ *     12/06/00  [rcg]
+ *              - Ensure daemon terminates on signal, by making listen
+ *                file descriptor non-blocking and calling select()
+ *                before accept(), since accept() does not return on
+ *                signals on some platforms.
+ *
  *     11/03/00  [rcg]
  *              - Ignore SIGCHLD on non-BSD systems.
  *              - HUP signal now closes and reopens log.
@@ -14,6 +34,9 @@
  *     10/10/00  [rcg]
  *              - Delete duplicate '#include <sys/wait.h>'
  *
+ *     10/13/00  [rcg]
+ *              - Added new TLS options (to avoid getopt errors).
+ *
  *     09/29/00  [rcg]
  *              - Revert signals to default in child (avoids conflict
  *                with BSDI authentication).
@@ -90,7 +113,18 @@ main ( argc, argv )
 int argc;
 char *argv[];
 {
-    return qpopper ( argc, argv );
+    if ( argc >= 2 && ( strncmp ( argv[1], "-v",  2 ) == 0 ||
+                        strncmp ( argv[1], "--v", 3 ) == 0   ) )
+    {
+        printf ( "%s%.*s%s version %s (non-standalone)\n",
+                 QPOP_NAME,
+                 (strlen(BANNERSFX)>0 ? 1 : 0), " ",
+                 BANNERSFX,
+                 VERSION );
+        return 0;
+    }
+    else
+        return qpopper ( argc, argv );
 }
 
 #else /* STANDALONE */
@@ -104,11 +138,22 @@ char *argv[];
 
 #define BAD_ADDR ( (unsigned long) -1 )
 
+/*
+ * Be careful using TRACE in an 'if' statement!
+ */
 #define TRACE if ( debug ) logit
 
 
 /*
- * Prototypes
+ * System prototypes (functions often missing prototypes in
+ * system header files)
+ */
+void bzero();
+pid_t wait3();
+
+
+/*
+ * Local prototypes
  */
 int     main ( int argc, char *argv[] );
 void    msg      ( WHENCE, const char *format, ... );
@@ -119,13 +164,16 @@ char   *sys_err_str ( void );
 int     reaper  ( SIGPARAM );
 int     hupit   ( SIGPARAM );
 int     cleanup ( SIGPARAM );
+void    roll_it ( void );
+void    motherforker ( int newsockfd, int sockfd );
 
 
 /*
  * Globals
  */
 char           *pname       = NULL;
-volatile BOOL   clean       =  0;
+volatile BOOL   bClean      = FALSE;
+volatile BOOL   bRollover   = FALSE;
 char          **Qargv       = NULL;
 int             Qargc       = 0;
 BOOL            Qargv_alloc = FALSE;
@@ -142,14 +190,11 @@ FILE           *err_out     = NULL;
  * can be specified as argv[1], e.g., 'popper 110'.
  */
 int
-main ( argc, argv )
-int argc;
-char *argv[];
+main ( int argc, char *argv[] )
 {
     int                 sockfd      = -1;
     int                 newsockfd   = -1;
     int                 clilen      =  0;
-    int                 childpid    =  0;
     int                 i           =  0;
     int                 rslt        =  0;
     struct sockaddr_in  cli_addr;
@@ -157,8 +202,22 @@ char *argv[];
     char               *ptr         = NULL;
     unsigned short      port        = SERV_TCP_PORT;
     unsigned long       addr        = INADDR_ANY;
+    fd_set              fdset_templ;
+    fd_set              fdset_read;
+    int                 fd_flags    = 0;
 
 
+    if ( argc >= 2 && ( strncmp ( argv[1], "-v",  2 ) == 0 ||
+                        strncmp ( argv[1], "--v", 3 ) == 0   ) )
+    {
+        printf ( "%s%.*s%s version %s (standalone)\n",
+                 QPOP_NAME,
+                 ( strlen(BANNERSFX) > 0 ? 1 : 0 ), " ",
+                 BANNERSFX,
+                 VERSION );
+        return 0;
+    }
+
     err_out = msg_out  = fopen ( "/dev/null", "w+" ); /* until we get set up */
 
     /*
@@ -253,13 +312,13 @@ char *argv[];
 #ifdef SYSLOG42
     openlog ( pname, 0 );
 #else
-    openlog ( pname, POP_LOGOPTS, POP_FACILITY );
+    openlog ( pname, POP_LOGOPTS, /*LOG_DAEMON*/ POP_FACILITY );
 #endif
 
     /*
      * See if debug or trace options specified
      */
-    i = getopt ( Qargc, Qargv, "b:BcdD:e:f:kK:L:p:RsSt:T:u" );
+    i = getopt ( Qargc, Qargv, "b:BcCdD:e:f:FkK:l:L:p:RsSt:T:uUvy:" );
     while ( i != EOF )
     {
         switch ( i )
@@ -270,19 +329,20 @@ char *argv[];
 
             case 't':
                 debug = TRUE;
-                trace_name = optarg;
+                trace_name = strdup ( optarg );
                 trace_file = fopen ( optarg, "a" );
                 if ( trace_file == NULL )
                     err_dump ( HERE, "Unable to open trace file \"%s\"", optarg );
-                TRACE ( trace_file, POP_DEBUG, HERE, "Opened trace file as %d",
-                        fileno(trace_file) );
+                TRACE ( trace_file, POP_DEBUG, HERE,
+                        "Opened trace file \"%s\" as %d",
+                        trace_name, fileno(trace_file) );
                 break;
 
             default:
                 break;
         }
 
-    i = getopt ( Qargc, Qargv, "b:BcdD:e:f:kK:L:p:RsSt:T:u" );
+        i = getopt ( Qargc, Qargv, "b:BcCdD:e:f:FkK:l:L:p:RsSt:T:uUvy:" );
     }
     optind = 1; /* reset for pop_init */
 
@@ -418,7 +478,7 @@ char *argv[];
     sockfd = socket ( AF_INET, SOCK_STREAM, 0 );
     if ( sockfd < 0 )
         err_dump ( HERE, "Can't open stream socket" );
-    TRACE ( trace_file, POP_DEBUG, HERE, "opened stream socket; sockfd = %d", socket );
+    TRACE ( trace_file, POP_DEBUG, HERE, "opened stream socket; sockfd = %d", sockfd );
  
     i = 1;
     rslt = setsockopt ( sockfd, SOL_SOCKET, SO_REUSEADDR, 
@@ -435,7 +495,7 @@ char *argv[];
                     "%s: Server: setsockopt(SO_DEBUG) failed", pname );
     }
 
-    TRACE ( trace_file, POP_DEBUG, HERE, "set stream socket options; sockfd = %d", socket );
+    TRACE ( trace_file, POP_DEBUG, HERE, "set stream socket options; sockfd = %d", sockfd );
 
     bzero ( (char *) &serv_addr, sizeof(serv_addr) );
     serv_addr.sin_family      = AF_INET;
@@ -458,7 +518,9 @@ char *argv[];
                        ntohs     ( serv_addr.sin_port ) );
     }
 
-    TRACE ( trace_file, POP_DEBUG, HERE, "did bind on stream socket; sockfd = %d", socket );
+    TRACE ( trace_file, POP_DEBUG, HERE,
+            "did bind on stream socket; sockfd = %d",
+            sockfd );
 
 
     /*
@@ -472,6 +534,25 @@ char *argv[];
 
     listen ( sockfd, 5 );
 
+    /*
+     * Set file descriptor to be non-blocking in case there isn't really a
+     * connection available if the select succeeds.  This avoids us
+     * blocking there.
+     */
+    fd_flags = fcntl ( sockfd, F_GETFL, 0 );
+    rslt     = fcntl ( sockfd, F_SETFL, O_NONBLOCK | fd_flags );
+    if ( rslt == -1 )
+        err_dump ( HERE, "Unable to set sockfd(%d) to be non-blocking",
+                   sockfd );
+    TRACE ( trace_file, POP_DEBUG, HERE, "set fd %d non-blocking (%#x)",
+            sockfd, fcntl ( sockfd, F_GETFL, 0 ) );
+
+    /*
+     * Build the template descriptor set
+     */
+    memset ( &fdset_templ, 0, sizeof(fdset_templ) );
+    FD_SET ( sockfd, &fdset_templ );
+
 #ifdef    BSD
     signal ( SIGCHLD, VOIDSTAR reaper  );
 #else /* not BSD */
@@ -483,13 +564,43 @@ char *argv[];
 
     while ( TRUE ) 
     {
-        if ( clean )
+        if ( bClean )
         {
             msg   ( HERE, "cleaning up and exiting normally" );
             close ( sockfd );
+            sockfd = -1;
+            if ( trace_file != NULL )
+            {
+                fclose ( trace_file );
+                trace_file = NULL;
+            }
             exit  ( 0 );
         }
 
+        if ( bRollover )
+        {
+            roll_it();
+            bRollover = FALSE;
+            signal ( SIGHUP,  VOIDSTAR hupit   );
+        }
+
+        /*
+         * Copy the template descriptor set into our read set (since
+         * select() modifies the fd set we pass in).
+         */
+        fdset_read = fdset_templ;
+
+        /*
+         * Check for a new connection with select() before calling accept(),
+         * since accept() does not return on signals on some platforms.
+         */
+        rslt = select ( sockfd + 1, &fdset_read, NULL, NULL, NULL );
+        if ( rslt == -1 && errno != EINTR )
+            err_dump ( HERE, "select() error" );
+
+        if ( rslt == 0 || ( FD_ISSET(sockfd, &fdset_read) == FALSE ) )
+            err_dump ( HERE, "unexpected select() result" ); 
+
         clilen    = sizeof(cli_addr);
         newsockfd = accept ( sockfd, (struct sockaddr *) &cli_addr, &clilen );
 
@@ -499,59 +610,13 @@ char *argv[];
                 inet_ntoa ( cli_addr.sin_addr ),
                 ntohs     ( cli_addr.sin_port ) );
 
-        if ( newsockfd < 0 )
-        {
-            if ( errno == EINTR )
-                continue;
-            err_dump ( HERE, "accept() error" );
-        }
-
-        TRACE ( trace_file, POP_DEBUG, HERE, "new connection; fd=%d", newsockfd );
-
-#ifndef _DEBUG
-        childpid = fork();
-        if ( childpid < 0 )
-            err_dump ( HERE, "fork() error" );
-        
-        else if ( childpid == 0 )
-        {
-            TRACE ( trace_file, POP_DEBUG, HERE, "new child for connection" );
-
-            /*
-             * Children should not trap signals
-             */
-            signal ( SIGCHLD, SIG_DFL );
-            signal ( SIGTERM, SIG_DFL );
-            signal ( SIGHUP,  SIG_DFL );
-
-            close ( sockfd );
-#endif /* not _DEBUG */
-
-            dup2    ( newsockfd, 0 );
-            dup2    ( newsockfd, 1 );
-            dup2    ( newsockfd, 2 );
-            close   ( newsockfd    );
-            qpopper ( Qargc, Qargv );
-            
-#ifdef _DEBUG
-            close  ( sockfd );
-#endif /* not _DEBUG */
-
-            TRACE ( trace_file, POP_DEBUG, HERE, "exiting after Qpopper returned" );
-            if ( Qargv_alloc )
-                free ( Qargv );
-            if ( trace_file != NULL )
-                fclose ( trace_file );
-
-            exit ( 0 );
-
-#ifndef _DEBUG
-        }
-
-        TRACE ( trace_file, POP_DEBUG, HERE, "forked() for new connection; pid=%d",
-                childpid );
-        close ( newsockfd );
-#endif /* not _DEBUG */
+        if ( newsockfd > 0 )
+            motherforker ( newsockfd, sockfd ); 
+        else
+            if ( errno != EINTR && errno != EWOULDBLOCK 
+                                && errno != EPROTO
+                                && errno != ECONNABORTED )
+                err_msg ( HERE, "accept() error" );
 
     } /* main loop */
 
@@ -559,6 +624,9 @@ char *argv[];
 }
 
 
+/*
+ * Logs a message
+ */
 void
 msg ( WHENCE, const char *format, ... )
 {
@@ -594,6 +662,9 @@ msg ( WHENCE, const char *format, ... )
 }
 
 
+/*
+ * Logs a message, dumps core, terminates.
+ */
 void
 err_dump ( WHENCE, const char *format, ... )
 {
@@ -621,8 +692,8 @@ err_dump ( WHENCE, const char *format, ... )
     left  -= ( iChunk > 0 ? iChunk : left );
     len   += ( iChunk > 0 ? iChunk : left );
 
-    iChunk = Qsnprintf ( msg_buf + len, left, ": %s [%d]",
-                         STRERROR(errno), errno );
+    iChunk = Qsnprintf ( msg_buf + len, left, ": %s [%s:%d]",
+                         sys_err_str(), fn, ln );
     left  -= ( iChunk > 0 ? iChunk : left );
     len   += ( iChunk > 0 ? iChunk : left );
 
@@ -632,9 +703,16 @@ err_dump ( WHENCE, const char *format, ... )
     logit   ( trace_file, POP_PRIORITY, fn, ln, "%s", msg_buf );
 
     if ( Qargv_alloc )
+    {
         free ( Qargv );
+        Qargv = NULL;
+    }
+
     if ( trace_file != NULL )
+    {
         fclose ( trace_file );
+        trace_file = NULL;
+    }
 
     fflush ( err_out );
 
@@ -643,6 +721,9 @@ err_dump ( WHENCE, const char *format, ... )
 }
 
 
+/*
+ * Logs a message, with error information automatically appended.
+ */
 void
 err_msg ( WHENCE, const char *format, ... )
 {
@@ -673,6 +754,11 @@ err_msg ( WHENCE, const char *format, ... )
 
     va_end   ( ap );
 
+    iChunk = Qsnprintf ( msg_buf + len, left, ": %s [%s:%d]",
+                         sys_err_str(), fn, ln );
+    left  -= ( iChunk > 0 ? iChunk : left );
+    len   += ( iChunk > 0 ? iChunk : left );
+
     fprintf ( err_out, "%s\n", msg_buf );
     logit   ( trace_file, POP_PRIORITY, fn, ln, "%s", msg_buf );
     fflush  ( err_out );
@@ -693,10 +779,10 @@ sys_err_str()
     if ( errno != 0 )
     {
         if ( errno > 0 && errno < sys_nerr )
-            Qsnprintf ( msgstr, sizeof(msgstr), " [(%d) %s]", 
+            Qsnprintf ( msgstr, sizeof(msgstr), "(%d) %s", 
                         errno, sys_errlist [ errno ] );
         else
-            Qsnprintf ( msgstr, sizeof(msgstr), " [errno = %d]", errno );
+            Qsnprintf ( msgstr, sizeof(msgstr), "errno = %d", errno );
 
         msgstr [ sizeof(msgstr) -1 ] = '\0';
     }
@@ -721,7 +807,9 @@ reaper ( SIGPARAM )
         child_pid = wait3 ( &stts, WNOHANG, (struct rusage *) 0 );
         TRACE  ( trace_file, POP_DEBUG, HERE, 
                  "...wait3() returned %d; error: %s (%d) ",
-                 child_pid, STRERROR(errno), errno );
+                 child_pid,
+                 ( child_pid > 0 ? "" : STRERROR(errno) ), 
+                 ( child_pid > 0 ?  0 : errno ) );
     }
         while  ( child_pid > 0 );
 
@@ -747,7 +835,7 @@ cleanup ( SIGPARAM )
                  child_pid, errno );
     }
 
-    clean = TRUE;
+    bClean = TRUE;
     return 0;
 }
 
@@ -756,6 +844,16 @@ int
 hupit ( SIGPARAM )
 {
     TRACE  ( trace_file, POP_DEBUG, HERE, "hupit signal handler called" );
+    signal ( SIGHUP, SIG_IGN );
+    bRollover = TRUE;
+    return 0;
+}
+
+
+void
+roll_it ( void )
+{
+    TRACE  ( trace_file, POP_DEBUG, HERE, "rolling over log..." );
 
     closelog();
 
@@ -765,8 +863,115 @@ hupit ( SIGPARAM )
     openlog ( pname, POP_LOGOPTS, LOG_DAEMON /*POP_FACILITY*/ );
 #endif
 
-    signal ( SIGHUP, VOIDSTAR hupit );
-    return 0;
+    if ( trace_file != NULL && trace_name != NULL )
+    {
+        TRACE  ( trace_file, POP_DEBUG, HERE, "rolling over trace file..." );
+        fflush ( trace_file );
+        fclose ( trace_file );
+
+        trace_file = fopen ( trace_name, "a" );
+        if ( trace_file == NULL )
+            err_dump ( HERE, "Unable to open trace file '%s'", trace_name );
+
+        TRACE ( trace_file, POP_DEBUG, HERE, "Opened trace file \"%s\" as %d",
+                trace_name, fileno(trace_file) );
+    }
+
+}
+
+
+/*
+ * Handles new client connection
+ */
+void
+motherforker ( int newsockfd, int sockfd )
+{
+    int     childpid    = 0;
+    int     fd_flags    = 0;
+    int     rslt        = 0;
+
+
+    TRACE ( trace_file, POP_DEBUG, HERE, "new connection; fd=%d", newsockfd );
+
+#ifndef _DEBUG
+    childpid = fork();
+    if ( childpid < 0 )
+        err_dump ( HERE, "fork() error" );
+    
+    else if ( childpid == 0 )
+    { /* I'm the child */
+        TRACE ( trace_file, POP_DEBUG, HERE, "new child for connection" );
+
+        /*
+         * Children should not trap signals
+         */
+        signal ( SIGCHLD, SIG_DFL );
+        signal ( SIGTERM, SIG_DFL );
+        signal ( SIGHUP,  SIG_DFL );
+
+        /*
+         * We don't need sockfd
+         */
+        close ( sockfd );
+        sockfd = -1;
+
+#endif /* not _DEBUG */
+
+        /*
+         * Make sure we pass a blocking socket to Qpopper
+         */
+        fd_flags = fcntl ( newsockfd, F_GETFL, 0 );
+        TRACE ( trace_file, POP_DEBUG, HERE, "newsockfd (%d) flags: %#x",
+                 newsockfd, fd_flags );
+        if ( fd_flags & O_NONBLOCK )
+        {
+            
+            rslt = fcntl ( newsockfd, F_SETFL, fd_flags - O_NONBLOCK );
+            if ( rslt == -1 )
+                err_dump ( HERE, "Unable to set newsockfd (%d) to be blocking",
+                           newsockfd );
+            TRACE ( trace_file, POP_DEBUG, HERE, "set fd %d blocking (%#x)",
+                    newsockfd, fcntl ( newsockfd, F_GETFL, 0 ) );
+        }
+
+        dup2    ( newsockfd, 0 );
+        dup2    ( newsockfd, 1 );
+        dup2    ( newsockfd, 2 );
+        close   ( newsockfd    );
+        newsockfd = -1;
+        qpopper ( Qargc, Qargv );
+            
+#ifdef _DEBUG
+        close  ( sockfd );
+        sockfd = -1;
+#endif /* not _DEBUG */
+
+        TRACE ( trace_file, POP_DEBUG, HERE, "exiting after Qpopper returned" );
+
+        if ( Qargv_alloc )
+        {
+            free ( Qargv );
+            Qargv = NULL;
+        }
+
+        if ( trace_file != NULL )
+        {
+            fclose ( trace_file );
+            trace_file = NULL;
+        }
+
+        _exit ( 0 );
+
+#ifndef _DEBUG
+    } /* I'm the child */
+    else
+    { /* I'm the parent */
+        TRACE ( trace_file, POP_DEBUG, HERE, "forked() for new connection; pid=%d",
+                childpid );
+        close ( newsockfd );
+        newsockfd = -1;
+    } /* I'm the parent */
+#endif /* not _DEBUG */
 }
 
 
