@@ -6,13 +6,37 @@
  
  
 /*
- * Copyright (c) 2000 Qualcomm Incorporated.  All rights reserved.
+ * Copyright (c) 2001 Qualcomm Incorporated.  All rights reserved.
  * The file License.txt specifies the terms for use, modification,
  * and redistribution.
  *
  *
  * Revisions:
  *
+ * 04/22/01  [rcg]
+ *          - pop_write_chunk() now correctly handles chunks longer
+ *            than the output buffer size.
+ *
+ * 02/23/01  [RCG]
+ *          - Debug octet checking only done if TRACE_MSG_BODY defined.
+ *
+ * 01/30/01  [RCG]
+ *          - NOW CHECKING p->bUpdate_status_hdrs instead of NO_STATUS.
+ *
+ * 01/25/01  [rcg]
+ *          - Implemented chunky_writes run-time option.
+ *          - Added error checking when sending data to client.
+ *          - Eliminated redundant calls on pop_write with no data.
+ *          - Checking p->bNo_mime.
+ *
+ * 01/24/01  [rcg]
+ *          - When DEBUG set, trace elapsed time and verify octet and
+ *            line counts.
+ *
+ * 10/12/00  [rcg]
+ *          - Fitted LGL's patches for tls support.
+ *          - Added pop_write_fmt() to send formatted output to client.
+ *
  * 08/23/00  [rcg]
  *          - Check for SIGPIP/SIGHUP more frequently.
  *
@@ -24,16 +48,16 @@
  * 03/01/00  [rcg]
  *          -  Now calling msg_ptr to adjust for hidden msgs and check range.
  *
- * 02/10/00  [RG]
+ * 02/10/00  [rcg]
  *          -  Don't trace message body unless TRACE_MSG_BODY defined.
  *
- * 01/18/00  [RG]
+ * 01/18/00  [rcg]
  *          -  Put octet count in OK response to RETR to appease
  *             certain poorly-written clients.
  *
  * 12/22/98  [RG]
  *  3/20/98  [PY]
- *        
+ * 
  */
 
 #include "config.h"
@@ -60,8 +84,25 @@
 #endif /* HAVE_INDEX */
 
 #include "popper.h"
+#include "pop_tls.h"
 #include "mmangle/mime.h"
 #include "mmangle/mangle.h"
+#include "snprintf.c"
+
+
+#ifdef    TRACE_MSG_BODY
+#  define TRACING_BODY  1
+#else
+#  define TRACING_BODY  0
+#endif /* TRACE_MSG_BODY */
+
+
+static void pop_write_now   ( POP *p, char *buffer, int length );
+static void pop_write_chunk ( POP *p, char *buffer, int length );
+
+
+static long total_octets_sent = 0;
+
 
 /* 
   Expects input one line at a time
@@ -157,40 +198,49 @@ static void *topper_init ( int            lines_to_output,
                            char           tp,
                            unsigned long  bl )
 {
-  topper_global.lines_to_output = lines_to_output;
-  topper_global.out_state       = out_state;
-  topper_global.in_header       = 1;
-  topper_global.last_char       = '\0';
-  topper_global.is_top          = tp;
-  topper_global.body_lines      = bl;
-  return ( &topper_global );
+    topper_global.lines_to_output = lines_to_output;
+    topper_global.out_state       = out_state;
+    topper_global.in_header       = 1;
+    topper_global.last_char       = '\0';
+    topper_global.is_top          = tp;
+    topper_global.body_lines      = bl;
+    return ( &topper_global );
 }
 
 
 static void topper_input ( void *topper_state, char *input, long len )
 {
-    char          *i, *i_end = input + len, *buf;
-    struct topper *t_state = (struct topper *)topper_state;
-    char          *crlf = "\r\n", dot = '.';
+    char          *i,
+                  *i_end = input + len,
+                  *buf;
+    struct topper *t_state = (struct topper *) topper_state;
+    char           dot     = '.';
+
+
+    if ( input == NULL )
+        return;
 
-    if(input == NULL) return;
     buf = input;
-    for(i = input; i < i_end; t_state->last_char = *i, i++) {
-        if(*i == '.' && t_state->last_char == '\n') {
-            fwrite(buf, 1, i - buf + 1, (FILE *)t_state->out_state);
-            fwrite(&dot, 1, 1, (FILE *)t_state->out_state);
+    for ( i = input;
+          i < i_end;
+          t_state->last_char = *i, i++ ) {
+        if ( *i == '.' && t_state->last_char == '\n' ) {
+            pop_write ( (POP *) t_state->out_state,  buf, i-buf+1 );
+            pop_write ( (POP *) t_state->out_state, &dot, 1       );
             buf = i+1;
-        } else if(*i == '\n') {
-            fwrite(buf, 1, i - buf, (FILE *)t_state->out_state);
-            fwrite(crlf, 1, 2, (FILE *)t_state->out_state);
+        } else
+        if ( *i == '\n' ) {
+            pop_write ( (POP *) t_state->out_state, buf,    i-buf );
+            pop_write ( (POP *) t_state->out_state, "\r\n", 2     );
             buf = i+1;
-            if(!t_state->in_header && t_state->is_top)
+            if ( !t_state->in_header && t_state->is_top )
                 t_state->lines_to_output--;
         } 
     }
-    fwrite(buf, 1, i - buf, (FILE *)t_state->out_state);
+    if ( i != buf )
+        pop_write ( (POP *) t_state->out_state, buf, i-buf );
 }
-  
+
 
 
 /* 
@@ -209,14 +259,21 @@ POP     *   p;
     void                   *hm_state, *t_state;
     MimeParsePtr            mimeParse = NULL;
     ManglerStateType        manglerState;
-    char                    is_top = 0;  /* Set to 1 if "top" */
+    BOOL                    is_top = FALSE;  /* Set to TRUE if "top" */
     int                     do_mangle; /* Contain the param number which contain
                                           mangle command */
     long                    bufLen = 0;
+    char                   *pLine  = NULL;
+    time_t                  timer       = time(0);
+    double                  sand        = 0;
+    int                     check_lines = 0;
+    int                     check_body  = 0;
+    int                     check_bytes = 0;
+    long                    check_total = total_octets_sent;
 
 
     /*  
-     * Convert the first parameter into an integer, it's the mesage num 
+     * Convert the first parameter into an integer; it's the mesage num 
      */
     msg_num = atoi ( p->pop_parm [ 1 ] );
 
@@ -258,19 +315,18 @@ POP     *   p;
                       ) 
                     )
                   ? 3 : 0;
-        is_top = 1;
+        is_top = TRUE;
     } /* top */
     else { /* retr */
         /* 
          * NO_STATUS does not dirty the mailspool if a status is changed 
          */
-#ifndef NO_STATUS
+
         /*  
          * Assume that a RETR (retrieve) command was issued 
          */
-        if ( mp->retr_flag != TRUE )
+        if ( p->bUpdate_status_hdrs && mp->retr_flag != TRUE )
             p->dirty = 1;
-#endif /* NO_STATUS */
 
         /*  
          * Flag the message as retreived 
@@ -285,6 +341,15 @@ POP     *   p;
                     ) ? 2 : 0;
     } /* retr */
 
+    /*
+     * Handle "-no-mime" user.  If this command didn't specify mangle
+     * but the user name ends in "-no-mime", we pretend mangle was used
+     * with default parameters (TEXT=
+     */
+    if ( do_mangle == 0 && p->bNo_mime ) {
+        do_mangle = 3;
+        p->pop_parm[3] = "X-MANGLE(TEXT=PLAIN)";
+    }
 
     /* 
      * Set up the output pipeline
@@ -293,7 +358,7 @@ POP     *   p;
      */
     t_state = topper_init (
                             top_lines,
-                            (void *) p->output,   /* Stream to output on */
+                            (void *) p,   /* Passed to pop_write_x */
                             is_top,
                             mp->body_lines - 1
                             );
@@ -331,12 +396,12 @@ POP     *   p;
     /*  
      * Position to the start of the message 
      */
-    (void)fseek ( p->drop, mp->offset, 0 );
+    (void) fseek ( p->drop, mp->offset, 0 );
 
     /*  
      * Skip the first line (the sendmail "From" or MMDF line) 
      */
-    (void)fgets ( buffer, MAXMSGLINELEN, p->drop );
+    (void) fgets ( buffer, MAXMSGLINELEN, p->drop );
 
     /* Some poorly-written clients (reported to include Netscape Messenger)
      * expect the octet count to be in the OK response to RETR, which is of
@@ -355,45 +420,66 @@ POP     *   p;
     /* 
      * Send the headers 
      */
-    while ( hangup == FALSE && fgets ( buffer, MAXMSGLINELEN, p->drop ) ) {
-        DEBUG_LOG3 ( p, "HDR (hangup=%d; errno=%d) *** %s",
-                     hangup, errno, buffer );
+    pLine = fgets ( buffer, MAXMSGLINELEN, p->drop );
+    while ( hangup == FALSE && pLine != NULL ) {
+        bufLen = strlen ( pLine );
+        if ( TRACING_BODY ) {
+            if ( strncasecmp ( pLine, "Content-Length:", 15 ) != 0 )
+                check_bytes += bufLen + 1; /* add one for CR */
+            check_lines++;
+        }
+        DEBUG_LOG4 ( p, "HDR (len=%ld; hangup=%d; errno=%d) *** %.50s",
+                     bufLen, hangup, errno, buffer );
         if ( strcmp(buffer,"\n") == 0 ) {
-            header_mucker_input ( hm_state, buffer, strlen(buffer) );
+            header_mucker_input ( hm_state, buffer, bufLen );
             break;
         }
-        header_mucker_input ( hm_state, buffer, strlen(buffer) );
-    }
+        header_mucker_input ( hm_state, buffer, bufLen );
+        pLine = fgets ( buffer, MAXMSGLINELEN, p->drop );
+    } /* while loop */
 
     /* 
      * Send the body 
      */
     ((struct topper *)t_state)->in_header = 0;
     if ( is_top ) {
+        pLine = fgets ( buffer, MAXMSGLINELEN, p->drop );
         while ( hangup == FALSE                           &&
-                fgets( buffer, MAXMSGLINELEN, p->drop )   && 
+                pLine != NULL                             && 
                ((struct topper*)t_state)->lines_to_output && 
                MSG_LINES-- ) {
-            bufLen = strlen(buffer);
-#ifdef TRACE_MSG_BODY
-            DEBUG_LOG4 ( p, "BDY (hangup=%d; errno=%d) *** %d(%d)[%li] %.256s", 
+        bufLen = strlen ( pLine );
+        if ( DEBUGGING ) {
+            check_bytes += bufLen + 1; /* add one for CR */
+            check_lines++;
+            check_body++;
+        }
+        if ( TRACING_BODY ) {
+            DEBUG_LOG6 ( p, "BDY (hangup=%d; errno=%d) *** %ld(%d)[%li] %.256s", 
                          hangup, errno, MSG_LINES, top_lines, bufLen, buffer );
-#endif /* TRACE_MSG_BODY */
-            header_mucker_input ( hm_state, buffer, bufLen );
         }
-    }
+        header_mucker_input ( hm_state, buffer, bufLen );
+        pLine = fgets ( buffer, MAXMSGLINELEN, p->drop );
+        } /* while loop */
+    } /* is_top */
     else { /* not TOP */
-        while ( hangup == FALSE                          &&
-                fgets ( buffer, MAXMSGLINELEN, p->drop ) &&
-                MSG_LINES-- ) {
-            bufLen = strlen(buffer);
-#ifdef TRACE_MSG_BODY
-            DEBUG_LOG3 ( p, "*** %d[%li] %.256s", 
+        pLine = fgets ( buffer, MAXMSGLINELEN, p->drop );
+        while ( hangup == FALSE && pLine != NULL && MSG_LINES-- ) {
+            bufLen = strlen ( pLine );
+            if ( DEBUGGING ) {
+                check_bytes += bufLen + 1; /* add one for CR */
+                check_lines++;
+                check_body++;
+            }
+            if ( TRACING_BODY ) {
+            DEBUG_LOG3 ( p, "BDY %ld[%li] %.256s", 
                          MSG_LINES, bufLen, buffer );
-#endif /* TRACE_MSG_BODY */
+            }
             header_mucker_input ( hm_state, buffer, bufLen );
-        }
-    }
+            pLine = fgets ( buffer, MAXMSGLINELEN, p->drop );
+        } /* while loop */
+    } /* not TOP */
+
     if ( do_mangle ) {
         FreeMangleInfo ( &manglerState.rqInfo );
         /* 
@@ -409,13 +495,39 @@ POP     *   p;
      */
     if ( topper_global.last_char != '\n' ){
         DEBUG_LOG0 ( p,"*** adding trailing CRLF" );
-        (void)fputs ( "\r\n", p->output );
+        pop_write ( p, "\r\n", 2 );
     }
 
-    (void)fputs  ( ".\r\n", p->output );
-    (void)fflush ( p->output );
+    pop_write_line  ( p, "." );
+    pop_write_flush ( p );
     DEBUG_LOG0   ( p, "---- sent termination sequence ----" );
 
+    if ( TRACING_BODY && DEBUGGING ) {
+        int expected_lines = mp->lines + 1; /* for X-UIDL */
+
+        check_total = total_octets_sent - check_total;
+        sand = difftime ( time(0), timer );
+        DEBUG_LOG3 ( p, "#elapsed time: %f seconds for message %d (%ld total octets)",
+                     sand, mp->number, check_total );
+
+        check_body++; /* separator counts as body */
+        check_lines++; /* don't forget X-UIDL */
+        check_bytes += strlen("X-UIDL: ") + strlen(mp->uidl_str) + 1;
+
+        if ( expected_lines != check_lines )
+            pop_log ( p, POP_NOTICE, HERE,
+                      "Incorrect line count (expected %d; found %d) msg %d in %s",
+                      expected_lines, check_lines, mp->number, p->drop_name );
+        if ( mp->body_lines != check_body )
+            pop_log ( p, POP_NOTICE, HERE,
+                      "Incorrect body lines (expected %d; found %d) msg %d in %s",
+                      mp->body_lines, check_body, mp->number, p->drop_name );
+        if ( mp->length != check_bytes )
+            pop_log ( p, POP_NOTICE, HERE,
+                      "Incorrect octet count (expected %ld; sent %d) msg %d in %s",
+                      mp->length, check_bytes, mp->number, p->drop_name );
+    }
+
     return ( POP_SUCCESS );
 }
 
@@ -439,27 +551,226 @@ char        *   buffer;
      * Byte stuff lines that begin with the termination octet 
      */
     if ( *buffer == POP_TERMINATE ) 
-        (void) fputc ( POP_TERMINATE, p->output );
+        pop_write ( p, POP_TERMINATE_S, 1 );
 
     /*  
      * Terminate the string at a <NL> if one exists in the buffer
      */
     bp = index ( buffer, NEWLINE );
     if ( bp != NULL ) 
-        *bp = 0;
+        *bp = '\0';
 
     /*  
      * Send the line to the client 
      */
-    (void) fputs ( buffer, p->output );
-
-    /*  
-     * Put a <CR><NL> if a newline was removed from the buffer 
-     */
-    if ( bp != NULL ) 
-        (void) fputs ( "\r\n", p->output );
+    pop_write_line ( p, buffer );
 
     return POP_SUCCESS;
 }
 
 
+
+/*
+ *  Write a null-terminated string to client with proper line ending. 
+ *  Does NOT dot-stuff, so this is for commands, not messages.
+ */
+void
+pop_write_line ( p, buffer )
+POP  *p;
+char *buffer;
+{
+    pop_write ( p, buffer, strlen(buffer) );
+    pop_write ( p, "\r\n", 2 );
+}
+
+
+
+
+/*
+ *  write a buffer to the client (may be pooled with later writes)
+ */
+void
+pop_write ( POP *p, char *buffer, int length )
+{
+#ifdef    _DEBUG
+    DEBUG_LOG4 ( p, "#chunky_writes=%d; tls_started=%d; %s %d bytes",
+                 (int) p->chunky_writes, p->tls_started,
+                 ( p->chunky_writes == ChunkyAlways
+                   || ( p->chunky_writes == ChunkyTLS
+                        && p->tls_started ) ) ? "buffering" : "sending",
+                 length );
+#endif /* _DEBUG */
+
+    if ( p->chunky_writes == ChunkyAlways
+         || ( p->chunky_writes == ChunkyTLS && p->tls_started ) )
+        pop_write_chunk ( p, buffer, length );
+    else
+        pop_write_now   ( p, buffer, length );
+}
+
+
+
+
+/* 
+ *  Write formatted text to the client.
+ */
+
+static char msgbuf      [ MAXLINELEN * 2 ];
+
+void
+pop_write_fmt ( POP         *p,         /* the source of all things */
+                const char  *format,    /* format string to log */
+                ... )                   /* parameters for format string */
+{
+    va_list          ap;
+    size_t           iLeft     = 0;
+    int              iChunk    = 0;
+
+
+    /* 
+     * Get the information into msgbuf 
+     */
+    va_start ( ap, format );
+
+    iLeft  = sizeof ( msgbuf ) -3; /* Allow for CRLF NULL at end */
+    iChunk = Qvsnprintf ( msgbuf, iLeft, format, ap );
+
+    va_end ( ap );
+
+    if ( iChunk == -1 ) {
+        /* 
+         * We blew out the format buffer.
+         */
+        pop_log ( p, POP_PRIORITY, HERE,
+                  "Buffer size exceeded formatting msg: %s", 
+                  format );
+        iChunk = strlen ( msgbuf ); /* send what we have */
+    } /* iChunk == -1 */
+
+    /*
+     * Now send it.
+     */
+    pop_write ( p, msgbuf, iChunk );
+    return;
+}
+
+
+
+
+/*
+ *  Flush the output that might be buffered to client
+ */
+void
+pop_write_flush ( POP *p )
+{
+    int rslt = 0;
+
+
+    if ( p->nOutBufUsed > 0 ) {
+        pop_write_now ( p, p->pcOutBuf, p->nOutBufUsed );
+        p->nOutBufUsed = 0;
+    }
+
+    if ( p->tls_started ) {
+        rslt = pop_tls_flush ( p->tls_context );
+    } else {
+        rslt = fflush ( p->output );
+    }
+
+    if ( rslt == EOF ) {
+        if ( p->tls_started )
+            pop_log ( p, POP_NOTICE, HERE, "Error flushing data to client" );
+        else {
+            int e = ferror ( p->output );
+            pop_log ( p, POP_NOTICE, HERE,
+                      "I/O error flushing output to client %s at %s [%s]: "
+                      "%s (%d)",
+                      p->user, p->client, p->ipaddr, STRERROR(e), e );
+        }
+    } /* flush failed */
+#ifdef    _DEBUG
+    else
+        DEBUG_LOG0 ( p, "#flushed output to client" );
+#endif /* _DEBUG */
+}
+
+
+
+
+/*
+ * send data to client without extra buffering
+ */
+static void
+pop_write_now ( POP *p, char *buffer, int length )
+{
+    int rslt = 0;
+
+
+    if ( p->tls_started ) {
+        rslt = pop_tls_write ( p->tls_context, buffer, length );
+    } else {
+        rslt = fwrite ( buffer, 1, length,  p->output );
+    }
+
+    if ( rslt == length ) {
+        total_octets_sent += rslt;
+#ifdef    _DEBUG
+        DEBUG_LOG2 ( p, "#wrote %d (out of %d) octets to client",
+                     rslt, length );
+#endif /* _DEBUG */
+        }
+    else { /* write failed */
+        if ( p->tls_started )
+            pop_log ( p, POP_NOTICE, HERE, "Error writing to client" );
+        else {
+            int e = ferror ( p->output );
+            pop_log ( p, POP_NOTICE, HERE,
+                    "I/O error writing to client %s at %s [%s]: %s (%d)",
+                    p->user, p->client, p->ipaddr, STRERROR(e), e );
+        }
+    } /* write failed */
+}
+
+
+
+
+/*
+ *  Add a buffer to the chunk of data to be sent to the client.
+ *  Data gets sent when chunk fills up, or pop-write_flush is
+ *  called.
+ */
+static void
+pop_write_chunk ( POP *p, char *buffer, int length )
+{
+    int nCopied = 0;
+    int nToCopy = 0;
+
+
+#ifdef    _DEBUG
+    DEBUG_LOG3 ( p, "#adding %d bytes to %d (out of %d max) currently buffered",
+                 length, p->nOutBufUsed, OUT_BUF_SIZE );
+#endif /* _DEBUG */
+
+    while ( length > 0 ) {
+       if ( p->nOutBufUsed + length > OUT_BUF_SIZE ) {
+           /*  new stuff won't fit in output buffer, flush old stuff */
+           pop_write_flush ( p );
+       }
+
+       nToCopy = OUT_BUF_SIZE - p->nOutBufUsed;
+       if ( length < nToCopy )
+            nToCopy = length;
+
+       bcopy ( buffer + nCopied,
+               &( p->pcOutBuf [ p->nOutBufUsed ] ),
+               nToCopy );
+       
+       p->nOutBufUsed += nToCopy;
+       nCopied        += nToCopy;
+       length         -= nToCopy;
+    } /* nCopied < length */
+}
+
+
+
+
