@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2000 QUALCOMM Incorporated.  All rights reserved.
+ * Copyright (c) 2001 QUALCOMM Incorporated.  All rights reserved.
  * The file license.txt specifies the terms for use, modification, or
  * redistribution.
  *
@@ -21,7 +21,7 @@
  */
 
 
-#include <config.h>
+#include "config.h"
 #include <stdio.h>
 #include <sys/types.h>
 
